/* NOTES:
- This is a GNU flex file (version 2.5.4).
- See example main files at the end
*/

%{
#include <string>
#include <list>
#include "my_parser/my_parser_includes.h"
#include "my_parser/my_parser.h"
#include <iostream>
#include <sstream>

// -----------------------------------------------------------------

void my_parser_error (std::string s);
std::string my_parser_error_string;

#if defined(MY_SCANNER_DEBUG) || defined(MY_PARSER_DEBUG)
std::string my_parser_debug_string;
#endif

bool my_parser_reached_end_of_file;
%}

/* Track line number */
%option yylineno

/* Make a case-insensitive scanner. */
%option case-insensitive

/* Set the name prefix so we can have multiple lexers in the same program. */
%option prefix="my_parser_"

/* Just have yywrap return 1. */
%option noyywrap

/* Necessary for Windows, because there is no isatty function, which is
 * generated for interactive execution. */
%option    never-interactive

LETTER              [A-Za-z]
UNDERSCORE          _
SPECIAL_CHARACTER   [!#%^&*()[\]\\|:'{}.<>/?$@`]
DIGIT               [0-9]
ZERO_DIGIT          0
NONZERO_DIGIT       [1-9]
DIGIT_SEQUENCE      {DIGIT}+
FOLLOWING_CHARACTER ({UNDERSCORE}|{LETTER}|{SPECIAL_CHARACTER}|{DIGIT})
IDENTIFIER          ({UNDERSCORE}|{LETTER}){FOLLOWING_CHARACTER}*
SIGN                [-+]
PERIOD              \.
DOTTED_DIGITS       ({DIGIT_SEQUENCE}{PERIOD}{DIGIT_SEQUENCE}?|{PERIOD}{DIGIT_SEQUENCE})
EXPONENT            [eE]{SIGN}?{DIGIT_SEQUENCE}
FLOAT_1             {SIGN}?{DIGIT_SEQUENCE}{EXPONENT}
FLOAT_2             {SIGN}?{DOTTED_DIGITS}{EXPONENT}?
NEWLINE             [\r\n]
WHITESPACE          [ \t]+

%s quoted

%%

<quoted><<EOF>> {
  my_parser_error( "Reached the end of the file while looking for the end of a quoted identifier" );
  yyterminate();
}

<INITIAL><<EOF>> {
#ifdef FT_SCANNER_DEBUG
  my_parser_debug_string += "Scanned the end of the file\n";
#endif // FT_SCANNER_DEBUG

  /* First time return an EOF token to the parser so that it can do final
   * checks. Second time abort */

  if (my_parser_reached_end_of_file)
  {
    yy_delete_buffer(YY_CURRENT_BUFFER);
    yyterminate();
  }
  else
  {
    my_parser_reached_end_of_file = true;
    return END_OF_FILE;
  }
}

<quoted>[^"]+ {
#ifdef MY_SCANNER_DEBUG
  my_parser_debug_string += "Scanned content of quoted identifier: ";
  my_parser_debug_string += my_parser_text;
  my_parser_debug_string += "\n";
#endif // MY_SCANNER_DEBUG
  my_parser_lval.text = new std::string(my_parser_text);
  return IDENTIFIER;
}

<quoted>\" {
#ifdef MY_SCANNER_DEBUG
  my_parser_debug_string += "Scanned the end of a quoted identifier.\n" ;
#endif // MY_SCANNER_DEBUG
  BEGIN(INITIAL);
}

\" {
#ifdef MY_SCANNER_DEBUG
  my_parser_debug_string += "Scanned the start of a quoted identifier.\n" ;
#endif // MY_SCANNER_DEBUG
  BEGIN(quoted);
}

{IDENTIFIER} {
#ifdef MY_SCANNER_DEBUG
  my_parser_debug_string += "Scanned an identifier: ";
  my_parser_debug_string += my_parser_text;
  my_parser_debug_string += "\n";
#endif // MY_SCANNER_DEBUG
  my_parser_lval.text = new std::string(my_parser_text);
  return IDENTIFIER;
}

({NONZERO_DIGIT}{DIGIT_SEQUENCE}?|{ZERO_DIGIT}+{NONZERO_DIGIT}{DIGIT_SEQUENCE}?) {
#ifdef MY_SCANNER_DEBUG
  my_parser_debug_string += "Scanned a natural: ";
  my_parser_debug_string += my_parser_text;
  my_parser_debug_string += "\n";
#endif // MY_SCANNER_DEBUG
  my_parser_lval.number = new float(atoi(my_parser_text));
  return NUMBER;
}

{DIGIT_SEQUENCE} {
#ifdef MY_SCANNER_DEBUG
  my_parser_debug_string += "Scanned a whole: ";
  my_parser_debug_string += my_parser_text;
  my_parser_debug_string += "\n";
#endif // MY_SCANNER_DEBUG
  my_parser_lval.number = new float(atoi(my_parser_text));
  return NUMBER;
}

{SIGN}?{DIGIT_SEQUENCE} {
#ifdef MY_SCANNER_DEBUG
  my_parser_debug_string += "Scanned an integer: ";
  my_parser_debug_string += my_parser_text;
  my_parser_debug_string += "\n";
#endif // MY_SCANNER_DEBUG
  my_parser_lval.number = new float(atoi(my_parser_text));
  return NUMBER;
}

({FLOAT_1}|{FLOAT_2}) {
#ifdef MY_SCANNER_DEBUG
  my_parser_debug_string += "Scanned a float: ";
  my_parser_debug_string += my_parser_text;
  my_parser_debug_string += "\n";
#endif // MY_SCANNER_DEBUG
  my_parser_lval.number = new float(atoi(my_parser_text));
  return NUMBER;
}

{NEWLINE} {
#ifdef MY_SCANNER_DEBUG
  my_parser_debug_string += "Scanned a newline.\n" ;
#endif // MY_SCANNER_DEBUG
  my_parser_lineno++;
}

{WHITESPACE} {
#ifdef MY_SCANNER_DEBUG
  my_parser_debug_string += "Scanned whitespace: ";
  my_parser_debug_string += my_parser_text;
  my_parser_debug_string += "\n";
#endif // MY_SCANNER_DEBUG
  /* do nothing */
}

"/*" {
  /* This doesn't currently pass the comment to the parser. We might want
     to do that later. */
#ifdef MY_SCANNER_DEBUG
  my_parser_debug_string += "Scanned a comment.\n" ;
#endif // MY_SCANNER_DEBUG
  register int c;

  for ( ; ; )
  {
    while ( (c = yyinput()) != '*' && c != EOF )
        ;    /* eat up text of comment */

    if ( c == '*' )
    {
      while ( (c = yyinput()) == '*' )
        ;
      if ( c == '/' )
        break;    /* found the end */
    }

    if ( c == EOF )
    {
      my_parser_error( "EOF in comment" );
      break;
    }
  }
}

. {
#ifdef MY_SCANNER_DEBUG
  my_parser_debug_string += "Scanned an unrecognized character: ";
  my_parser_debug_string += my_parser_text;
  my_parser_debug_string += "\n";
#endif // MY_SCANNER_DEBUG
  std::string s;
  s = "Invalid character: ";
  s += my_parser_text;
  my_parser_error(s.c_str());
}

%%

void my_parser_error(std::string error_string)
{
  std::ostringstream temp_string;
  temp_string << my_parser_lineno;

  // Generic yacc "syntax error" message isn't very useful
  if (error_string != "syntax error")
    my_parser_error_string += "SYNTAX ERROR (line " + temp_string.str() +
      "): " + error_string + "\n";
  else
    my_parser_error_string += "SYNTAX ERROR (line " + temp_string.str() +
      ")\n";
}

void my_parser_scanner_initialize()
{
  my_parser_error_string = "";
#if defined(MY_SCANNER_DEBUG) || defined(MY_PARSER_DEBUG)
  my_parser_debug_string = "";
#endif // defined(MY_SCANNER_DEBUG) || defined(MY_PARSER_DEBUG)
}

/*
Sample main file

#include <string>
#include <stdio.h>

extern FILE *my_scanner_in;
extern int my_parser_lex();

int main(int argc, char *argv[]) {
  my_scanner_in = fopen(argv[1], "r");
  while (my_parser_lex())
    ;
  fclose(my_scanner_in);

  return 0;
}
*/
